/********************************************************************
 * FILE: gps_sirf.c
 * DESCRIPTION:
 * Utility functions for initializing, reading, parsing binary
 * SiRF messages(Ref: SiRF binary protocol, ver2.4)
 *
 *  Created on: Apr 18, 2010
 *      Author: Rohit Pandita
 * University of Minnesota 
 * Aerospace Engineering and Mechanics 
 *  2011 Regents of the University of Minnesota. All rights reserved.
 *
 * $Id: gps_sirf.c 253 2011-01-20 22:01:35Z murch $
 ********************************************************************/

/* SiRF packet structure
 * <0xA0,0xA2> <packet length> <packet> <cksum> <0xB0,0xB3>
 * bytes: <2> <2> <variable_size> <2> <2>
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <termios.h>
#include <math.h>

#include "gps_sirf.h"
#include "../../utils/serial_mpc5200.h"




int init_gps( char *portname, int baudrate )
{
	int port;
	unsigned char cmdNmea2BinaryMode_B9600[] = "$PSRF100,0,9600,8,1,0*xx\r\n";
	unsigned char cmdNmea2BinaryMode_B57600[] = "$PSRF100,0,57600,8,1,0*xx\r\n";
	unsigned char cmdNmea2BinaryMode_B38400[] = "$PSRF100,0,38400,8,1,0*xx\r\n";
	unsigned char cmdNmea2BinaryMode_B115200[] = "$PSRF100,0,115200,8,1,0*xx\r\n";

	// get serial port handle
	cyg_io_lookup( portname, &ttl0_handle );

	// open serial port at default GPS baudrate 4800
	port = open_serial( portname, B4800 );
	if (port < 0)
		return port;

	//change GPS output to binary at given baudrate
	switch( baudrate ){
	case B9600:
		nmea_add_checksum( cmdNmea2BinaryMode_B9600 );
		write( port, cmdNmea2BinaryMode_B9600, sizeof(cmdNmea2BinaryMode_B9600) );
	case B38400:
		nmea_add_checksum( cmdNmea2BinaryMode_B38400 );
		write( port, cmdNmea2BinaryMode_B38400, sizeof(cmdNmea2BinaryMode_B38400) );
	case B57600:
		nmea_add_checksum( cmdNmea2BinaryMode_B57600 );
		write( port, cmdNmea2BinaryMode_B57600, sizeof(cmdNmea2BinaryMode_B57600) );
	case B115200:
		nmea_add_checksum( cmdNmea2BinaryMode_B115200 );
		write( port, cmdNmea2BinaryMode_B115200, sizeof(cmdNmea2BinaryMode_B115200) );
	}

	sleep(1);
	//write(port, cmdNmea2BinaryMode, sizeof(cmdNmea2BinaryMode));
	//sleep(1);
	//write(port, cmdNmea2BinaryMode, sizeof(cmdNmea2BinaryMode));
	close(port);

	// change serial port baudrate
	port = open_serial( portname, baudrate );

	return port;
}

void nmea_add_checksum( unsigned char *sentence )
{
	unsigned char sum = '\0';
	unsigned char c, *p = sentence;

	if (*p == '$' || *p == '!') {
		p++;
	}

	while (((c = *p) != '*') && (c != '\0')) {
		sum ^= c;
		p++;
	}

	*p++ = '*';

	snprintf(p, 5, "%02X\r\n", (unsigned)sum);

}

int read_gps(struct gps *gpspkt, int port)
{
	cyg_io_get_config (ttl0_handle, CYG_IO_GET_CONFIG_SERIAL_BUFFER_INFO,\
			&buff_info, &len);
	unsigned int bytesInBuffer = buff_info.rx_count;

	//	gpspkt->err_type = gps_nolock;

	if( bytesInBuffer > 0 )
		return RxSirfPacket( port, gpspkt, bytesInBuffer );
	else {
		gpspkt->err_type = got_invalid;
		return 0;
	}
}

int RxSirfPacket( int port, struct gps *pckt, unsigned int bytesinbuffer )
{
	unsigned char rxbuf[256];
	unsigned int SirfHeaderCheck=0, nbytes=0, count=0;
	unsigned short msgPacketsize, readAttempt=0;

	// locate SiRF binary packet header 0xA0A2
	while( (SirfHeaderCheck != 2) && (count < bytesinbuffer) ){

		while( 1 != read(port,&rxbuf[0],1) ); count++;

		if (rxbuf[0] == 0xa0){
			SirfHeaderCheck++;
			while(1 != read(port,&rxbuf[1],1)); count++;

			if (rxbuf[1] == 0xa2)
				SirfHeaderCheck++;
			else
				SirfHeaderCheck = 0;
		}
		else
			SirfHeaderCheck = 0;
	}

	if( count >= bytesinbuffer ){
		pckt->err_type = got_invalid;
		return(-1);
	}

	/*
	// find out size of the packet
	while( nbytes != 2 ) nbytes += read(port, rxbuf+2+nbytes, 2-nbytes);
	nbytes=0;
	 */

	// find out size of the packet
	if( 2 != read(port, rxbuf+2, 2) ){
		pckt->err_type = got_invalid;
		return(-1);
	}

	//
	//msgPacketsize =  getbeuw(rxbuf,2) + 4;		// including 4 byte tail, but excluding header(4 bytes)
	// Included by David Escobar
	msgPacketsize =  (getbeuw(rxbuf,2)&0x7FFF ) + 4; // including 4 byte tail, but excluding header(4 bytes)

#ifdef verbose
	fprintf( stderr, "\n <SiRF_drv>: packet of %d received.",msgPacketsize);
#endif

	// read packet into temp buffer
	while( (nbytes != msgPacketsize) && (readAttempt++ < 3) )
		nbytes += read(port, rxbuf+4+nbytes, msgPacketsize-nbytes);
	nbytes=0;

	if( readAttempt >= 3 ){
		pckt->err_type = got_invalid;
		return(-1);
	}

	if ( verifySirfCksum( rxbuf, msgPacketsize ) == TRUE ){
		//pckt->err_type = data_valid;
		parseSirfPacket( rxbuf, pckt );
		return(1);
	}
	else{
		pckt->err_type = checksum_err;
		return(-1);
	}

}

int verifySirfCksum(byte *buf, int packetsize)
{
	int i;
	// getbeuw : big endian unsigned word
	uint16_t calcsum=0, recvsum = getbeuw(buf,packetsize) & 0x7FFF;

	for ( i=0; i<packetsize-4; i++ ){
		calcsum += buf[i+4];
	}
	calcsum = calcsum & 0x7FFF;

	if ( calcsum == recvsum )
		return TRUE;
	else
		return FALSE;
}

void parseSirfPacket( byte *b, struct gps *pckt )
{
	byte *base = b+4;
	byte msgID = base[0];
	double speedOverGround, courseOverGround;

	switch ( msgID ){
	case 0x06: // msgID6: response to poll software version (msg 132)
		//	fprintf( 1, "SiRF firmware version: %s", base+1 );
#ifdef verbose
		fprintf( stderr, "<SiRF_drv>: Packet ID %d ", msgID);
#endif
		break;

	case 0x02: // msgID2: Measure Nav data out (ECEF)
		pckt->Xe = (double)getbesl(base,1);
		pckt->Ye = (double)getbesl(base,5);
		pckt->Ze = (double)getbesl(base,9);
		pckt->Ue = (double)getbesw(base,13) / 8;
		pckt->Ve = (double)getbesw(base,15) / 8;
		pckt->We = (double)getbesw(base,17) / 8;
		//mode = *(b+18)
		pckt->GPS_week = getbeuw(base,22);
		pckt->GPS_TOW = getbeul(base,24) / 100;
		pckt->SatVisible = base[28];
		// append time tag
		pckt->time = get_Time();
		break;


	case 0x29:	 //msgID41: Geodetic Nav Data
		if( getbeuw(base,1) != 0)
			pckt->err_type = gps_nolock;

		else{
			pckt->err_type = data_valid;

			pckt->UTC_year = getbeuw(base,11);
			pckt->UTC_month = base[13];
			pckt->UTC_day = base[14];
			pckt->UTC_hour = base[15];
			pckt->UTC_min = base[16];
			pckt->UTC_secs = getbeuw(base,17)*1e-03;

			pckt->lat = (double)getbesl(base,23)*1e-07;
			pckt->lon = (double)getbesl(base,27)*1e-07;
			pckt->alt = (double)getbesl(base,35)*1e-02;			// AMSL	meters
			speedOverGround = (double)getbeuw(base,40)*1e-02;		// m/s
			courseOverGround = D2R*(double)getbeuw(base,42)*1e-02;	// rad
			pckt->vd = -(double)getbesw(base,46)*1e-02;
			pckt->vn = cos(courseOverGround)*speedOverGround;
			pckt->ve = sin(courseOverGround)*speedOverGround;
			// append time tag
			pckt->time = get_Time();

			//			printf("\n Lat=%f,\t Lon=%f, vn=%f",pckt->lat,pckt->lon,pckt->vn);
		}
		break;

	case 0x1c: // msgID28: Carrier phase info (Nav Lib Measurement Data)
		/* not yet implemented */
		break;
	}

}

